import tensorflow as tf

from . import math as tk
from .utils import wrap_log_k
from .utils import log_bessel_recurrence


@wrap_log_k
def log_bessel_k(v, x):
    v = tk.abs(v)
    n = tk.round(v)
    u = v - n
    log_ku, log_kup1 = _log_bessel_ku(u, x)
    return log_bessel_recurrence(log_ku, log_kup1, u, n, x)[0]


def _log_bessel_ku(u, x, mask=None, max_iter=100, return_counter=False):
    """
    I.J. Thompson and A.R. Barnett,
    Modified Bessel function Iv(z) and Kv(z) and real order
    and complex argument, to selected accuracy,
    Computer Physics Communications, 47, 245-257 (1987).
    """

    def cond(si, ri, i, bi, ci, di, fp, fi, gi, hi):
        nonzero_update = tk.abs(di * hi) > tol * tk.abs(si)
        if mask is not None:
            nonzero_update &= mask
        return tf.reduce_any(nonzero_update)

    def body(si, ri, i, bi, ci, di, fp, fi, gi, hi):
        j = i + 1.
        aj = tf.square(j - 0.5) - tf.square(u)
        bj = 2. * (x + j)

        cj = 1. / (bj - aj * ci)
        dj = di * (bj * cj - 1.)
        rj = ri + dj

        fj = (bi * fi - fp) / aj
        gj = gi * aj / j
        hj = hi + fj * gj
        sj = si + dj * hj
        return sj, rj, j, bj, cj, dj, fi, fj, gj, hj

    x = tf.convert_to_tensor(x)
    u = tf.convert_to_tensor(u, x.dtype)
    if mask is not None:
        mask = tf.convert_to_tensor(mask, tf.bool)

    tol = tk.epsilon(x.dtype)

    i = tf.cast(1., x.dtype)
    a1 = 0.25 - tf.square(u)
    b1 = 2. * x + 2.

    c1 = 1. / b1
    d1 = 1. / b1
    r1 = d1

    f0 = tf.zeros_like(u * x)
    f1 = tf.ones_like(u * x)
    g1 = a1
    h1 = f1 * g1
    s1 = 1. + d1 * h1

    init = s1, r1, i, b1, c1, d1, f0, f1, g1, h1
    sn, rn, counter, *_ = tf.while_loop(
        cond, body, init, maximum_iterations=max_iter,
    )
    log_ku = 0.5 * tk.log(0.5 * tk.pi / x) - x - tk.log(sn)
    log_kup1 = log_ku + tk.log((0.5 + u + x - a1 * rn) / x)

    if return_counter:
        return log_ku, log_kup1, counter
    return log_ku, log_kup1



import tensorflow as tf

from . import math as tk


def get_deriv_func(func, i=0):
    def deriv(*args):
        with tf.GradientTape() as g:
            g.watch(args)
            f = func(*args)
        return g.gradient(f, args[i])
    return deriv


def extend(func, x0, dx):

    def cond(x, d, f1):
        return tf.reduce_any(~tf.equal(dx, 0.) & (f1 > 0))

    def body(x, d, f1):
        f1 = func(x + d)
        x = tf.where(f1 > 0., x + d, x)
        d = tf.where(f1 > 0., 2. * d, d)
        return x, d, f1

    init = x0, dx, tf.ones_like(x0)
    x, d, _ = tf.while_loop(cond, body, init)
    return x, x + d


def find_zero(func, x0, x1, tol, max_iter):

    def cond(x0, x1, f0, f1):
        return tf.reduce_any(~tf.equal(x0, x1) & (tk.abs(f0) > 1.))

    def body(x0, x1, f0, f1):
        x_shrink = x0 + 0.5 * (x1 - x0)
        f_shrink = func(x_shrink)

        dx = - f0 / deriv(x0) / (x1 - x0)
        dx_in_range = (0.0 < dx) & (dx < 0.5)
        x_newton = tf.where(dx_in_range, x0 + dx * (x1 - x0), x0)
        f_newton = func(x_newton)

        c_shrink = f_shrink < 0
        c_newton = f_newton < 0
        x0_new = tf.where(c_shrink, x_shrink, tf.where(c_newton, x_newton, x0))
        f0_new = tf.where(c_shrink, f_shrink, tf.where(c_newton, f_newton, f0))
        x1_new = tf.where(c_shrink, x1, tf.where(c_newton, x_shrink, x_newton))
        f1_new = tf.where(c_shrink, f1, tf.where(c_newton, f_shrink, f_newton))
        return x0_new, x1_new, f0_new, f1_new

    deriv = get_deriv_func(func)
    f0 = func(x0)
    f1 = func(x1)
    init = x0, x1, f0, f1
    return tf.while_loop(cond, body, init, maximum_iterations=max_iter)[0]


def log_bessel_recurrence(log_ku, log_kup1, u, n, x, mask=None):

    def cond(ki, kj, ui, ni):
        should_update = ni > 0
        if mask is not None:
            should_update &= mask
        return tf.reduce_any(should_update)

    def body(ki, kj, ui, ni):
        uj = ui + 1.
        nj = ni - 1
        kk = tk.log_add_exp(ki, kj + tk.log(2. * uj / x))
        k0 = tf.where(ni > 0, kj, ki)
        k1 = tf.where(ni > 0, kk, kj)
        return k0, k1, uj, nj

    init = log_ku, log_kup1, u, n
    return tf.while_loop(cond, body, init)[:2]


def wrap_log_k(native_log_k):

    def wraped_log_k(v, x, name=None):

        @tf.custom_gradient
        def _log_K_custom_gradient(v, x):
            return native_log_k(v, x), _log_K_grad

        def _log_K_grad(u):
            logkv = _log_K_custom_gradient(v, x)
            logkvm1 = _log_K_custom_gradient(v - 1., x)
            dlogkvdx = - v / x - tk.exp(logkvm1 - logkv)
            return None, u * dlogkvdx

        with tf.name_scope(name or 'bessel_K'):
            x = tf.convert_to_tensor(x)
            v = tf.convert_to_tensor(v, x.dtype)
            return _log_K_custom_gradient(v, x)

    return wraped_log_k



import tensorflow as tf
import numpy as np


pi = np.pi
inf = np.inf
nan = np.nan


is_nan = tf.math.is_nan
is_finite = tf.math.is_finite


minimum = tf.math.minimum
maximum = tf.math.maximum

sign = tf.math.sign
abs = tf.math.abs
round = tf.math.round

reciprocal = tf.math.reciprocal
square = tf.math.square
sqrt = tf.math.sqrt
pow = tf.math.pow

log = tf.math.log
log1p = tf.math.log1p
xlogy = tf.math.xlogy

exp = tf.math.exp
expm1 = tf.math.expm1

sin = tf.math.sin
cos = tf.math.cos
tan = tf.math.tan

sinh = tf.math.sinh
cosh = tf.math.cosh
tanh = tf.math.tanh

log_gamma = tf.math.lgamma

sum = tf.math.reduce_sum
log_sum_exp = tf.math.reduce_logsumexp


def epsilon(dtype):
    return np.finfo(dtype.as_numpy_dtype).eps


def gamma(x):
    return exp(log_gamma(x))


@tf.custom_gradient
def log_sinh(x):
    def grad(upstream):
        return upstream / tanh(x)
    return tf.where(x < 20., log(sinh(x)), x - np.log(2.)), grad


@tf.custom_gradient
def log_cosh(x):
    def grad(upstream):
        return upstream * tanh(x)
    return x + log1p(expm1(-2 * x) / 2), grad


def sinc(x):
    pix = np.pi * x
    return tf.where(tf.equal(x, 0.), tf.cast(1., x.dtype), sin(pix) / pix)


def sinhc(x):
    return tf.where(tf.equal(x, 0.), tf.cast(1., x.dtype), sinh(x) / x)


def log_add_exp(x, y, sign=None):
    larger = maximum(x, y)
    if sign is None:
        sign = 1.
    return larger + log(exp(x - larger) + sign * exp(y - larger))


def log_sub_exp(x, y):
    larger = maximum(x, y)
    return larger + log(exp(x - larger) - exp(y - larger))



import tensorflow as tf

from . import math as tk
from .utils import get_deriv_func, extend, find_zero


def log_bessel_k(v, x, n=0, m=0, name=None):

    @tf.custom_gradient
    def _custom_gradient(v, x, n, m):
        return _log_bessel_k(v, x, n, m), lambda u: _grad(n, m, u)

    def _K_grad(n, m, u):
        logkv = _custom_gradient(v, x, n, m)
        dlogkvdv = tk.exp(_custom_gradient(v, x, n + 1, m) - logkv)
        dlogkvdx = tk.exp(_custom_gradient(v, x, n, m + 1) - logkv)
        dlogkvdx = tf.where(tf.equal(m % 2, 0), -dlogkvdx, dlogkvdx)
        return u * dlogkvdv, u * dlogkvdx, None, None

    with tf.name_scope(name or 'bessel_K_tk2'):
        x = tf.convert_to_tensor(x)
        v = tf.convert_to_tensor(v, x.dtype)
        return _custom_gradient(v, x, n, m)


def _log_bessel_k(v, x, n=0, m=0,
                  mask=None, dt0=1., tol=1., max_iter=10, bins=128):

    def func(t):
        out = tf.where(
            tf.equal(n % 2, 0), tk.log_cosh(v * t), tk.log_sinh(v * t),
        )
        out -= x * tk.cosh(t)
        out += tf.where(n > 0, nf * tk.log(t), tf.cast(0., x.dtype))
        out += tf.where(m > 0, mf * tk.log_cosh(t), tf.cast(0., x.dtype))
        return out

    def func_mth(t):
        return func(t) - th

    x = tf.convert_to_tensor(x)
    v = tf.convert_to_tensor(v, x.dtype)
    if mask is None:
        mask = tf.ones_like(v * x, tf.bool)
    else:
        mask = tf.convert_to_tensor(mask, tf.bool)
    mask &= (x > 0.) & tf.where(n > 0, ~tf.equal(v, 0.), tf.cast(1, tf.bool))

    eps = tk.epsilon(x.dtype)
    zero = tf.zeros_like(v * x)
    dt0 = dt0 * tf.ones_like(v * x)
    deriv = get_deriv_func(func)

    nf = tf.cast(n, x.dtype)
    mf = tf.cast(m, x.dtype)

    out = tf.cast(tk.nan, x.dtype) # x < 0.
    out = tf.where(tf.equal(x, 0.), tf.cast(tk.inf, x.dtype), out)

    positive_peak = ~tf.equal(n, 0) | (tf.square(v) + mf > x)
    if mask is not None:
        positive_peak &= mask
    dt = tf.where(positive_peak, dt0, zero)
    ts, te = extend(deriv, zero, dt)
    tp = find_zero(deriv, te, ts, tol, max_iter)
    th = func(tp) + tk.log(eps)

    tpm = tk.maximum(tp - bins * eps, 0.)
    zero_exists = func_mth(zero) < 0.
    if mask is not None:
        zero_exists &= mask
    t0 = find_zero(func_mth, zero, tpm, tol, max_iter)

    tpp = tk.maximum(tp + bins * eps, tp * (1. + bins * eps))
    zero_exists = func_mth(tpp) > 0.
    if mask is not None:
        zero_exists &= mask
    dt = tf.where(zero_exists, dt0, zero)
    ts, te = extend(func_mth, tpp, dt)
    t1 = find_zero(func_mth, te, ts, tol, max_iter)

    t = tf.linspace(t0, t1, 2 * bins + 1, axis=0)[1:-1:2]
    h = (t1 - t0) / bins
    sum_eft = tk.sum(tk.exp(func(t) - func(tp)), axis=0)
    return tf.where(mask, func(tp) + tk.log(h * sum_eft), out)
